#ifndef __PPSSUMM_LOCAL_ALGO__
#define __PPSSUMM_LOCAL_ALGO__
/* 
 * algorithms used by ppssumm which are not specific to ppssumm
 *
 * Tseq(f, n, Os) Append a series of integers (f,f+n] to sequence Os
 * 
 * reorder( I1 , I2 , Or1 , Or2 ) Reorder the elements of sequence (I1,I2]
 * according to the positions read from sequence (Or1,Or2]
 * 
 * sort_order( I1 , I2 , Os ) stable sort of sequence (I1 , I2] also
 * adding a sequence of original positions to Os (ie. *(os+0) = 4 indicates the 
 * element in I1+0 after the sort was originally in position I1+4).
 *
 * find_all ( I1 , I2 , Os , Pred)
 * Find all elements of sequence (I1 , I2] for which pred(*In) returns true.  Add positions (relative to I1) of matching elements to sequence Os.  Matching is by Pred(
 * 
 * find_all ( I1 , I2 , J1 , J2 , Os )
 * Find all elements of sequence (I1 , I2] which are matched by an element in (J1 , J2]. Add positions (relative to I1) of matching elements to sequence Os.
 * 
 * find_all ( I1 , I2 , I1::value_type , Os )
 * Find all elements of sequence (I1 , I2] which are equal to the value provided. Add positions (relative to I1) of matching elements to sequence Os.
 * 
 * equal_ranges( I1 , I2 , Os ) 
 * Add pair<diff , diff> to Os which give the offsets of (I1 , I2] with equal values.
 * (eg. { 0 2 4 6 6 6 7 8 8 } Has equal ranges (3,6] and (7,9] ).
 *
 */
#include <algorithm>
#include <functional>
#include <iterator>
#include <utility>
#include <vector>

/* 
 * Turn a number + count into an iterator sequence
 */
 
template< typename _T , typename _Oi >
_Oi Tseq(_T _first , _T _many , _Oi _seq) {

	for( ; _many > 0 ; _many--) 
		*_seq++ = _first++;
	return _seq;
	
}

template< typename _Ii , typename _Io >
void reorder( _Ii _first , _Ii _last , _Io _ofirst , _Io _olast ) {

	/* 
	 * Move sequence element _Ii[n] to position _Ii[*_Io]
	 * for each _Ii and _Io
	 */
 	if (_first == _last || _ofirst == _olast ) return;

	typedef typename iterator_traits<_Ii>::difference_type idiff;
	typedef typename iterator_traits<_Ii>::value_type itype;
	typedef typename iterator_traits<_Io>::value_type otype;
	

	idiff diff = _last - _first;
	//typename iterator_traits<_Ii>::value_type tmp[ static_cast<unsigned int>(diff) ];
	itype tmp[ diff ];

	for(idiff i = 0 
		; i < diff  && _ofirst != _olast 
		; i++ , _ofirst++ 
	) {
		*(tmp+i) =   *(_first + ( static_cast<idiff>(*(_ofirst)) ) );
	}

	for(idiff i = 0 ; i < diff ; i++ ) *(_first+i) = *(tmp+i);
		
}

/*
template< typename _Ri , typename _Oi , typename _Pred >
void find_all( _Ri _first , _Ri _last , _Oi _order , _Pred _pred )
{
	// Like sort_order, collect the offsets of all elements in (_first,_last]
	// for which _Fn _comp is true
	//
	typename iterator_traits<_Ri>::difference_type i = 0;
	while (_first != _last) {
		if (_pred(*_first)) _order++ = i;
		i++;
		_first++;
	}
	
}
*/

template< typename _Ri , typename _Oi >
void find_all( _Ri _first1 , _Ri _last1 , _Ri _first2 , _Ri _last2 , _Oi _order )
{
	/* 
	 * Like sort_order, collect the offsets of all elements in (_first1,_last1]
	 * which equal an element in (_first2 , _last2)
	 */

	typename std::iterator_traits<_Ri>::difference_type i = 0;
	for ( ; _first1 != _last1 ; _first1++ , i++ ) {

		if ( std::find( _first2 , _last2 , *_first1 ) != _last2 ) {

			_order++ = i;

		}	
	}

	
}

template< typename _Ri , typename _Oi >
void find_all( _Ri _first , _Ri _last , typename iterator_traits<_Ri>::value_type  _value , _Oi _order )
{
	/* 
	 * Like sort_order, collect the offsets of all elements in (_first1,_last1]
	 * which equal an element in (_first2 , _last2)
	 */

	typename std::iterator_traits<_Ri>::difference_type i = 0;
	for ( ; _first != _last; _first++ , i++ ) {

		if (*_first == _value) _order++ = i;
	}

	
}


template< typename _Ri , typename _Oi > //, typename _Fn >
void sort_order( _Ri _first , _Ri _last , _Oi _order ) //, _Fn _comp )
{
	/* 
	 * Operation: stable_sort of sequence _first to _last
	 * and build a record of each elements original position relative to _first (=0)
	 * 
	 * Do this by building a temporary container of pairs and stablesort them on their .first element. 
	 * the cmp_pair comparason below.
	 * 
	 * Assuming position numbers n,m {0 .. N} assign new positions to 
	 * _order[n] = m where m be the original position of the element
	 * such that element[m] should be moved to position n for it
	 * to reach it's final sorted position.
	 *
	 * eg.  Input Sequence:  2 6 1 6 7
	 * 	Becomes       :  1 2 6 6 7
	 * 	giving _order :  2 0 1 3 4
	 *
	 * The whole point of this is I need the new index ordering to 
	 * re-order other sequences based on the sorting of another
	 *
	 * See the reorder() function below for how to do that.
	 */


	if (_first == _last) return;

	typedef typename iterator_traits<_Ri>::difference_type idiff;
	typedef typename iterator_traits<_Ri>::value_type itype;
	typedef std::pair< itype , idiff > element;

	typedef std::vector< element > container;


	/* 
	 * Using an array here caused a bus error when optimization
	 * was switched on.  Hence use of container
	 */
	container data;

	idiff pos=0;
	for (_Ri _tmp = _first ; _tmp != _last ; _tmp++ , pos++) 
	{
		data.push_back( std::make_pair( *_tmp , pos ) ) ;
	}

	std::stable_sort( data.begin() , data.end() );

	for ( typename container::const_iterator CI = data.begin()
		; CI != data.end()
		; CI++ , _first++ , _order++
	) {
		*_first = CI->first;
		*_order = CI->second;
	}

}


template< typename _Ri , typename _Oi > //, typename _Fn >
void equal_ranges( _Ri _first , _Ri _last  , _Oi _result)
{

	if (_first == _last) return;

	typedef std::pair< _Ri , _Ri > Eqrange;
	typedef typename std::iterator_traits<_Ri>::difference_type offset;
	typedef std::pair< offset , offset > offset_range;

	Eqrange eqrange = std::make_pair( _first , _last );

	while ( eqrange.first != _last ) {
		
		eqrange = std::equal_range( eqrange.first 
			, _last 
			, *(eqrange.first) 
		);

		if ( (eqrange.second - eqrange.first) 
		) {
			*_result++ = std::make_pair( eqrange.first - _first
				, eqrange.second - _first
			);
		} 

		eqrange.first = eqrange.second;
	
	}


}

template< typename _Ri , typename _Oi > 
void table_sort( _Ri _first , _Ri _last  , _Oi _order)
{
	
	/* scratch pad
	 *
	 * sort_order first column giving master row_order
	 * make a list of equal ranges.
	 *
	 * for each next column:
	 * 	reorder column based on master row_order
	 * 	stable_sort within equal ranges of previous column
	 * 	reorder range of master row_order defined by equal range using new stable_sort order.
	 * 	find equal ranges from new column within equal ranges of old column.
	 * 	
	 */

	typedef typename std::iterator_traits<_Ri>::value_type ColType;
	typedef typename std::iterator_traits<_Ri>::difference_type offset;
	typedef std::pair< offset , offset > offset_range;


	if (_first == _last) return;

	ColType _col = *_first;

	/* sort the first column, re-order others based on this sorting */

	std::vector<offset> row_order;
	sort_order( _col.begin() , _col.end() ,  back_inserter(row_order) );

	/* 
	 * extract equal ranges from first column. These are the ranges which will 
	 * require sorting in subsequent columns
	 */

	vector< offset_range > offsetranges;
	equal_ranges( _col.begin() , _col.end() , back_inserter(offsetranges) );
	
	/* Now the rest of the columns */
	_first++;
	while ( _first != _last ) {
		_col = *_first;

		/* reorder this column in accordance with row_roder */
		reorder( _col.begin() 
			, _col.end() 
			, row_order.begin() 
			, row_order.end() 
		);

		/*
		 * Now sort_order this column over the equal_range's of the previous column.
		 * reorder the master row_order in these equal_range's according to new sub_order.
		 */
		vector< offset_range > newranges;
		for ( typename vector< offset_range >::const_iterator CI = offsetranges.begin()
			; CI != offsetranges.end()
			; CI++
		){
			if ( _col.begin() + CI->first == _col.end()) continue;

			std::vector<unsigned int> sub_order;
			
			sort_order( _col.begin() + CI->first , _col.begin() + CI->second  , std::back_inserter(sub_order) );

			reorder( row_order.begin() + CI->first 
				, row_order.begin() + CI->second 
				, sub_order.begin() 
				, sub_order.end() 
			);

			/* 
			 * Find equal_range's of this column within the equal_ranges defined by previous columns
			 * Remeber, equal ranges in this column outside of these prior equal ranges have been sorted according
			 * to a previous column and so are sorted.
			 */
			
			equal_ranges( _col.begin() + CI->first ,  _col.begin() + CI->second , back_inserter(newranges) );
		}
		offsetranges = newranges;
		_first++;
	}
	
	/* Fill the new column ordering container */
	for (std::vector<unsigned int>::const_iterator CO = row_order.begin() 
		; CO != row_order.end() 
		; CO++ )
		_order++ = *CO;
}


#endif
